/* (C) Copyright 2008 Bill Buckels */

#define TRUE 1
#define FALSE 0

#include <io.h>
#include <poke.h>

/* a "limited" directory lister */
/* for files of a 3 character extension */
/* or files of any extension */
/* this will not work for files without extensions */

/*

This module uses the native C64 SYSIO library functions
in conjunction with a Commodore 64 ROM call to the
built-in LOAD routine to list the files on the
program disk.

This is similar to the BASIC 2 command sequence:

LOAD "$",8
LIST

In fact this module uses exactly the same routines
but compresses the directory listing retrieved by the
LOAD routine into a buffer starting at 6144.

It does so by walking-through the buffer and pulling the
filenames from between the quote delimiters in the
buffer.

Although it helps if one is somewhat familiar with
C64 BASIC and the C64 flavor of 6502 assembler as
well as the C64 Aztec C native library and in fact
the Commodore 64 memory map and how the Aztec C linker
is used to build this thing, this still provides a good
read even for those who don't or can't remember
if they do.

The following link will be generally helpful
when it comes to knowing a little more about
Commodore DOS :

http://en.wikipedia.org/wiki/Commodore_DOS


*/


/*

B-15. Function Name: LOAD
Purpose: Load RAM from device
Call address: $FFD5 (hex) 65493 (decimal)
Communication registers:A, X, Y
Preparatory routines: SETLFS, SETNAM
Error returns: 0,4,5,8,9, READST
Stack requirements: None
Registers affected: A, X, Y

Description: This routine LOADs data bytes from any input device di-
rectly into the memory of the Commodore 64. It can also be used for a
verify operation, comparing data from a device with the data already
in memory, while leaving the data stored in RAM unchanged.

The accumulator (.A) must be set to 0 for a LOAD operation, or 1 for a
verify, If the input device is OPENed with a secondary address (SA) of
0 the header information from the device is ignored. In this case, the
X and Y registers must contain the starting address for the load. If
the device is addressed with a secondary address of 1, then the data
is loaded into memory starting at the location specified by the
header. This routine returns the address of the highest RAM location
loaded.

Before this routine can be called, the KERNAL SETLFS, and SETNAM
routines must be called. NOTE: You can NOT LOAD from the keyboard (0),
RS-232 (2), or the screen (3).


How to Use:

Call the SETLFS, and SETNAM routines. If a relocated load is desired,
use the SETLFS routine to send a secondary address of 0.

Set the A register to 0 for load, 1 for verify.

If a relocated load is desired, the X and Y registers must be set to
the start address for the load.

Call the routine using the JSR instruction.

EXAMPLE:

         ;LOAD   A FILE FROM TAPE

          LDA #DEVICE1        ;SET DEVICE NUMBER
          LDX #FILENO         ;SET LOGICAL FILE NUMBER
          LDY CMD1            ;SET SECONDARY ADDRESS
          JSR SETLFS
          LDA #NAME1-NAME     ;LOAD A WITH NUMBER OF
                              ;CHARACTERS IN FILE NAME
          LDX #<NAME          ;LOAD X AND Y WITH ADDRESS OF
          LDY #>NAME          ;FILE NAME
          JSR SETNAM
          LDA #0              ;SET FLAG FOR A LOAD
          LDX #$FF            ;ALTERNATE START
          LDY #$FF
          JSR LOAD
          STX VARTAB          ;END OF LOAD
          STY VARTA B+1
          JMP START
  NAME    .BYT 'FILE NAME'
  NAME1   ;

*/


/*

B-28. Function Name: SETLFS
Purpose: Set up a logical file
Call address: $FFBA (hex) 65466 (decimal)
Communication registers: A, X, Y
Preparatory routines: None
Error returns: None
Stack requirements: 2
Registers affected: None

Description: This routine sets the logical file number, device
address, and secondary address (command number) for other KERNAL
routines.

The logical file number is used by the system as a key to the file
table created by the OPEN file routine. Device addresses can range
from 0 to 31. The following codes are used by the Commodore 64 to
stand for the CBM devices listed below:

ADDRESS  DEVICE
0  Keyboard
1  DatassetteTM
2  RS-232C device
3  CRT display
4  Serial bus printer
8  CBM serial bus disk drive


Device numbers 4 or greater automatically refer to devices on the
serial bus.

A command to the device is sent as a secondary address on the serial
bus after the device number is sent during the serial attention
handshaking sequence. If no secondary address is to be sent, the Y
index register should be set to 255.

How to Use:


Load the accumulator with the logical file number.
Load the X index register with the device number.
Load the Y index register with the command.
EXAMPLE:

   FOR LOGICAL FILE 32, DEVICE #4, AND NO COMMAND:
   LDA #32
   LDX #4
   LDY #255
   JSR SETLFS


*/




dlist(prg,ext)
char *prg, *ext;
{

    /* this function moves the names of any valid files
       to the beginng of our memory hole at 6144 and leaves the
       rest of our memory hole for other uses */

    /* best called at the start of a program.
       the file list can then be moved as needed to some other
       area of memory. I am overlaying the same area
       that I load the graphics font into so for a graphics program
       the font should be relocated before making this call
       or reloaded afterwards, after something is done with the listing. */
    int idx, found, c,d,e,f,p,r,g, ctr, first, base = 0;

    char E='*', X='*', T='*';

    /* args are either a wild card for all or a 3 character extension */
    if (ext[0] != '*') {
		E = ext[0];
		X = ext[1];
		T = ext[2];
	}


    dirtop = 6144; /* set the top of the listing in memory
                      to the base of the directory array.
                      this will be advanced for each file found */


	_dset(prg); /* seed the device listing with this
	                 program's name. assume that they
	                 have not pulled the disk out of the
	                 drive */

	first = FALSE; /* no disk name yet */
	found = FALSE; /* no names of any kind yet */
	ctr = 0;
	for (idx = 0; idx < 8192;idx++) { /* 8192 is an unlikely size for a directory */
	  c=dirptr[idx];
	  /* if a quote character is found it may be the beginning or end of
	     a file name so test further */
	  if (c == '\"') {
		  if (found==FALSE){
			  /* test for recognized file types and file names */
			  if (idx == 5) {
                  found = TRUE;
			  }
			  else {
				  p = dirptr[idx+19]; /* type of file */
				  r = dirptr[idx+20];
				  g = dirptr[idx+21];

				  if (p == 'P' && r == 'R' && g == 'G')found = TRUE; /* program */
				  if (p == 'S' && r == 'E' && g == 'Q')found = TRUE; /* sequential */
				  if (p == 'R' && r == 'E' && g == 'L')found = TRUE; /* relative */
				  if (p == 'U' && r == 'S' && g == 'R')found = TRUE; /* user specified */
				  if (p == 'D' && r == 'E' && g == 'L')found = TRUE; /* internal use */

				  if (found == TRUE) {
					 /* we are using the file extension as a directory
					    filter string. there is no telling what kind of trouble
					    we might have otherwise */
					 p = FALSE;
					 r = idx+1;
					 /* check for a valid file and pop it to the next position
					    in our directory list if we find it. */
					 for (g = 0; g < 13; g++) {
					    c = dirptr[r];
                        if (c== '\"')break;
                        if (c=='.') {
						   d = dirptr[r+1];
						   e = dirptr[r+2];
						   f = dirptr[r+3];
						   if (E == '*') {
						      p = TRUE;
						      break;
						   }
						   if (d == E  && e == X && f == T) {
							  p = TRUE;
							  break;
						   }
						}
					    r++;
					 }
					 if (p == TRUE) {
					 	p = base;
					 	r = idx+1;
					 	for (g = 0; g < 17; g++) {
							/* static length of 17 is stuffed with 0's first */
						 	dirptr[base] = 0; base++; dirtop++;
					 	}
					 	for (g = 0; g < 16; g++) {
						 	c = dirptr[r];
						 	if (c== '\"')break;
						 	dirptr[p] = c;
						 	p++;
						 	r++;
				     	}
					 	ctr++; /* we are only counting the valid files */
					 }
					 first = TRUE;
				  }
			  }
		  }
		  else {
			  found = FALSE;  /* turn-off file name */
			  first = TRUE;
		  }
	  }
	  else {
		  if(found != TRUE) {
			 /* the end of the listing is terminated with "BLOCKS FREE." */
			 if (c=='B') {
				 /* quick check on first 3 characters */
			     p = dirptr[idx+1];
			     r = dirptr[idx+2];
			     g = dirptr[idx+3];
			     if (p == 'L' && r == 'O' && g == 'C') {
			       if (strncmp("BLOCKS FREE.",(char *)&dirptr[idx],12)==0) {
					 break;
				   }

			   }
			 }
		  }

      }
	}

	return ctr;

}


/* use Aztec C's Commodore 64 SYSIO functions
   to set-up the directory lister then
   call a helper function to complete
   the command */
_dset(prg)
char *prg;
{
	struct fil_tab fp; /* static file table for lowlevel call */

	errno = 0;
	/* get the device that we have been loaded from */
	if (_get_dev(prg, &fp) < 0)
		return(-1);
    /* set logical file number, device address, and indicate
       that we will be using a secondary load address.
       in this case we will be using our 2K buffer below
       the graphics screen */
	_setlfs(2,fp.dev,0);
	/* the next line is reminiscent of BASIC */
	_setnam("$"); /* set name to directory */
    /* now call the rom routine to load the directory into
       our buffer */
    return _dlode();
}

/* yes virginia, there is inline assembly in Aztec C */
_dlode()
{
#asm

* rather than muck with passing stack args
* i kept it simple and hard-coded the relocatable
* address for the dir buffer which is $1800

LOAD equ $ffd5


* SET FLAG FOR A LOAD
	lda #0
* ALTERNATE START
    ldx #0
    ldy #$18
    jsr LOAD
    rts

#endasm
}



