/* Copyright (C) Bill Buckels 2008 */

/* sid chip sound playback routines */

/*

MEANINGS OF SOUND TERMS

ADSR     -- Attack/Decay/Sustain/Release
Attack   -- Rate sound rises to peak volume
Decay    -- Rate sound falls from peak volume to Sustain level
Sustain  -- Prolong note at certain volume
Release  -- Rate at which volume falls from Sustain level
Waveform -- "Shape" of sound wave
Pulse    -- Tone quality of Pulse Waveform


ADSR is encoded in 2 bytes

A - attack - high nibble (4 bits)
D - decay - low nibble (4 bits)

S - sustain - high nibble (4 bits)
R - release - low nibble (4 bits)

                        +
                       / \
                      /   \
                     /     \
   SUSTAIN LEVEL . ./. . . .+--------+
                   /                  \
                  /                    \
                 /                      \

                 |      |   |        |   |
                 |   A  | D |    S   | R |



These two 1-byte registers control the shape of the wave form
generated by the SID chip. Together with wave type (and pulse
settings if a pulse wave type), these registers define a musical
instrument for each of the three voices of the SID chip.

There are 3 sets of registers for each of the 3 voices offset by 7
bytes from each other.

Attack/Decay and Sustain/Release (ADSR) settings should always be
set in your program BEFORE the Waveform is set.


  +-----+------------------------+--------------------------------+
  |VALUE|ATTACK RATE (TIME/CYCLE)| DECAY/RELEASE RATE (TIME/CYCLE)|
  +-----+------------------------+--------------------------------+
  |  0  |           2 ms         |               6 ms             |
  |  1  |           8 ms         |              24 ms             |
  |  2  |          16 ms         |              48 ms             |
  |  3  |          24 ms         |              72 ms             |
  |  4  |          38 ms         |             114 ms             |
  |  5  |          56 ms         |             168 ms             |
  |  6  |          68 ms         |             204 ms             |
  |  7  |          80 ms         |             240 ms             |
  |  8  |         100 ms         |             300 ms             |
  |  9  |         250 ms         |             750 ms             |
  | 10  |         500 ms         |             1.5 s              |
  | 11  |         800 ms         |             2.4 s              |
  | 12  |           1 s          |               3 s              |
  | 13  |           3 s          |               9 s              |
  | 14  |           5 s          |              15 s              |
  | 15  |           8 s          |              24 s              |
  +-----+------------------------+--------------------------------+


+----------------------------------------------------------------------------+
|SETTING VOLUME -- SAME FOR ALL 3 VOICES                                     |
+--------------+---------+---------------------------------------------------+
|VOLUME CONTROL|POKE54296| Settings from 0 (off) to 15 (loudest)             |
+--------------+---------+---------------------------------------------------+
                              VOICE NUMBER 1
+--------------+---------+---------------------------------------------------+
|TO CONTROL    |POKE THIS|         FOLLOWED BY ONE OF THESE NUMBERS          |
|THIS SETTING: |NUMBER:  | (0 to 15 ... or ... 0 to 255 depending on range)  |
+--------------+---------+---------------------------------------------------+
|TO PLAY A NOTE|      C  | C#| D | D#| E | F | F#| G | G#| A | A#| B | C | C#|
|HIGH FREQUENCY|54273 34 | 36| 38| 40| 43| 45| 48| 51| 54| 57| 61| 64| 68| 72|
|LOW FREQUENCY |54272 75 | 85|126|200| 52|198|127| 97|111|172|126|188|149|169|
+--------------+---------+------------+------------+------------+------------+
|WAVEFORM      |  POKE   |  TRIANGLE  |  SAWTOOTH  |   PULSE    |   NOISE    |
|              |  54276  |     17     |     33     |     65     |    129     |
+--------------+---------+------------+------------+------------+------------+
|PULSE RATE (Pulse Waveform)                                                 |
|HI POLSE      |  54275  |   A value of 0 to 15  (for Pulse waveform only)   |
|LO POLSE      |  54274  |   A value of 0 to 255 (for Pulse waveform only)   |
+--------------+---------+------+------+------+------+-----+-----+-----+-----+
|ATTACK/       |  POKE   | ATK4 | ATK3 | ATK2 | ATK1 | DEC4| DEC3| DEC2| DEC1|
|       DECAY  |  54277  | 128  |  64  |  32  |  16  |  8  |  4  |  2  |  1  |
+--------------+---------+------+------+------+------+-----+-----+-----+-----+
|SUSTAIN/      |  POKE   | SUS4 | SUS3 | SUS2 | SUS1 | REL4| REL3| REL2| REL1|
|       RELEASE|  54278  | 128  |  64  |  32  |  16  |  8  |  4  |  2  |  1  |
+--------------+---------+------+------+------+------+-----+-----+-----+-----+
                              VOICE NUMBER 2
+--------------+---------+---------------------------------------------------+
|TO CONTROL    |POKE THIS|         FOLLOWED BY ONE OF THESE NUMBERS          |
|THIS SETTING: |NUMBER:  | (0 to 15 ... or ... 0 to 255 depending on range)  |
+--------------+---------+---------------------------------------------------+
|TO PLAY A NOTE|      C  | C#| D | D#| E | F | F#| G | G#| A | A#| B | C | C#|
|HIGH FREQUENCY|54280 34 | 36| 38| 40| 43| 45| 48| 51| 54| 57| 61| 64| 68| 72|
|LOW FREQUENCY |54279 75 | 85|126|200| 52|198|127| 97|111|172|126|188|149|169|
+--------------+---------+------------+------------+------------+------------+
|WAVEFORM      |  POKE   |  TRIANGLE  |  SAWTOOTH  |   PULSE    |   NOISE    |
|              |  54283  |     17     |     33     |     65     |    129     |
+--------------+---------+------------+------------+------------+------------+
|PULSE RATE (Pulse Waveform)                                                 |
|HI POLSE      |  54282  |   A value of 0 to 15  (for Pulse waveform only)   |
|LO POLSE      |  54281  |   A value of 0 to 255 (for Pulse waveform only)   |
+--------------+---------+------+------+------+------+-----+-----+-----+-----+
|ATTACK/       |  POKE   | ATK4 | ATK3 | ATK2 | ATK1 | DEC4| DEC3| DEC2| DEC1|
|       DECAY  |  54284  | 128  |  64  |  32  |  16  |  8  |  4  |  2  |  1  |
+--------------+---------+------+------+------+------+-----+-----+-----+-----+
|SUSTAIN/      |  POKE   | SUS4 | SUS3 | SUS2 | SUS1 | REL4| REL3| REL2| REL1|
|       RELEASE|  54285  | 128  |  64  |  32  |  16  |  8  |  4  |  2  |  1  |
+--------------+---------+------+------+------+------+-----+-----+-----+-----+
                              VOICE NUMBER 3
+--------------+---------+---------------------------------------------------+
|TO CONTROL    |POKE THIS|         FOLLOWED BY ONE OF THESE NUMBERS          |
|THIS SETTING: |NUMBER:  | (0 to 15 ... or ... 0 to 255 depending on range)  |
+--------------+---------+---------------------------------------------------+
|TO PLAY A NOTE|      C  | C#| D | D#| E | F | F#| G | G#| A | A#| B | C | C#|
|HIGH FREQUENCY|54287 34 | 36| 38| 40| 43| 45| 48| 51| 54| 57| 61| 64| 68| 72|
|LOW FREQUENCY |54286 75 | 85|126|200| 52|198|127| 97|111|172|126|188|149|169|
+--------------+---------+------------+------------+------------+------------+
|WAVEFORM      |  POKE   |  TRIANGLE  |  SAWTOOTH  |   PULSE    |   NOISE    |
|              |  54290  |     17     |     33     |     65     |    129     |
+--------------+---------+------------+------------+------------+------------+
|PULSE RATE (Pulse Waveform)                                                 |
|HI POLSE      |  54289  |   A value of 0 to 15  (for Pulse waveform only)   |
|LO POLSE      |  54288  |   A value of 0 to 255 (for Pulse waveform only)   |
+--------------+---------+------+------+------+------+-----+-----+-----+-----+
|ATTACK/       |  POKE   | ATK4 | ATK3 | ATK2 | ATK1 | DEC4| DEC3| DEC2| DEC1|
|       DECAY  |  54291  | 128  |  64  |  32  |  16  |  8  |  4  |  2  |  1  |
+--------------+---------+------+------+------+------+-----+-----+-----+-----+
|SUSTAIN/      |  POKE   | SUS4 | SUS3 | SUS2 | SUS1 | REL4| REL3| REL2| REL1|
|       RELEASE|  54292  | 128  |  64  |  32  |  16  |  8  |  4  |  2  |  1  |
+--------------+---------+------+------+------+------+-----+-----+-----+-----+

*/



#define SID 54272

char *_spoke = (char *)0;
char *_spress = (char *)203;


/* the following C64 SID frequencies represent
   general midi octaves -1 through 6 */
/* to find midi note 60 (mid C) the offset is 60 */
/* don't go by the note values in the table above...
   I have adjusted these for pitch. */

/* if the midi note in the SIB file is out of range
   no sound is played in the playback routine */

unsigned int _cfreq[95]={
  274,  291,  308,  326,  345,  366,  388,  411,  435,  461,  488,  517,
  549,  582,  616,  653,  691,  733,  776,  822,  871,  923,  977, 1035,
 1099, 1164, 1233, 1306, 1383, 1466, 1553, 1645, 1742, 1846, 1955, 2071,
 2198, 2328, 2466, 2613, 2767, 2932, 3106, 3290, 3485, 3692, 3911, 4143,
 4396, 4657, 4933, 5226, 5535, 5864, 6212, 6580, 6970, 7384, 7822, 8287,
 8793, 9315, 9867,10452,11071,11728,12424,13161,13941,14769,15645,16574,
17557,18600,19704,20874,22113,23426,24818,26292,27853,29508,31261,33118,
35084,37170,39378,41718,44196,46822,49606,52554,55676,58986,62492};


/* a pointer to the base of the SID chip registers */
char *_sid = (char *)SID;

/* frequency registers for all 3 SID voices */
int *_freq1 = (int *)SID;
int *_freq2 = (int *)(SID + 7);
int *_freq3 = (int *)(SID +14);

/* instrument settings array follows */
/* this idea is similar to BASIC 7's envelope on the C128 */

/* the following instruments are the combined values from
   the examples in the C64 user's and programmer's guides
   and the C128 programmer's guide.

   there are lots more sophisticated ways for designing
   instruments even on the C64 and these are just a starting
   point and must never be taken as more than that.

   although wave settings are used properly in this program
   the sound will only be as good as the emulator and the
   host computer that they are run-on.

   I am ignoring filter settings in all of this lot as
   well and really only the very basics of SID sound
   are covered in this program.

   OTOH it is pretty easy to understand for the most part,
   and I had no trouble writing it from the ground-up.

   */
char _inset[19][5] = {
65,     9,     0,  255, 0,  /* Piano       */
65,     9,     0,  0, 6, 	/* Piano2      */
17,    88,    89,  0, 0,	/* Violin      */
33,     9,    33,  0, 0,	/* Guitar      */
17,    96,     0,  0, 0, 	/* Flute       */
17,   148,    64,  0, 0,	/* Flute2      */
33,     9,     0,  0, 0, 	/* Harpsichord */
65,     9,     0,  0, 2, 	/* Harpsichord2*/
17,     9,     0,  0, 0, 	/* Xylophone   */
17,     9,     9,  0, 0,	/* Xylophone2  */
17,     0,   240,  0, 0, 	/* Organ       */
65,     9,   144,  0, 8, 	/* Organ2      */
17,     0,   240,  0, 0, 	/* Colliape    */
17,   102,     0,  0, 0, 	/* Accordian   */
33,   192,   192,  0, 0,	/* Accordian2  */
33,    96,     0,  0, 0, 	/* Trumpet     */
65,   137,    65,  0, 2, 	/* Trumpet2    */
129,    5,    80,  0, 0,	/* Drum        */
129,    5,    80,  0, 0};   /* custom - user specified values */


/* if the second and third voices are set then
   they accompany the first voice one octave higher
   and lower in "harmony" or on the same note in "unison"

   in "harmony" playback the voice2 and voice3 playback
   requires a longer duration than 1/10th of a second.

   this is admittedly an empirical decision that I have
   made on this since there is no real basis in musical
   theory for what I am doing in this lot.

   It's only bit-music and 3 voices and doesn't need to be
   overcomplicated with synopticons and such.

   */

int _hu = 0;
int _tempo = 812;

harmony()
{
	_hu = 12;
}

unison()
{
	_hu = 0;
}

/* tickvalue(0) resets tempo to default */
/* this is actually not tempo but the length of
   the delay which modifies duration.
   812 is an approximation for 1 tick at 18.2 ticks per second */

/* this value may vary slightly on the C64 and you may wish
   to adjust it for fine-tuning...
   or you may wish to speed-up or slow-down a note playback sequence
   (a song) so can speed-up the song by decreasing this value
   and conversely slow-down a song by increasing this value.
   keep in mind that I do absolutely no range checking for this.
   and an out of range composite of delay * tempo can occur */

tickvalue(tkval)
int tkval;
{
   _tempo = 812;

   if (tkval > 0) {
	   _tempo = tkval;

   }
}


/* duration is given in 18.2 ticks per second
   which is based on the IBM-PC 1Ch timer tick interrupt.

   this is a convenient form of time compression for
   several reasons. one of them is to conserve memory
   with a 1 byte duration value. another is that
   this allowed me to translate several songs that I already
   had in IBM format to a C64 equivalent without much effort. */

/* voices are set to valid instrument numbers */

play(song,voice1, voice2, voice3)
char *song;
int voice1, voice2, voice3;
{
   unsigned freq, note, dur, len;
   register idx;

   /* type of wave for all three voices */
   /*  triangle, sawtooth, pulse, noise */
   /*  17, 33, 65, 129 */
   /* these values are set by choosing an instrument */
   char wv1=0, wv2=0, wv3=0;

   for (idx = 0; idx < 25; idx++)_sid[idx] = 0; /* clear sound chip */

   /* make sure all voices are in the range of the instruments */
   if (voice1 < 0 || voice1 > 18)voice1 = 0;
   if (voice2 < 0 || voice2 > 18)voice2 = -1;
   if (voice3 < 0 || voice3 > 18)voice3 = -1;

   /* set-up voice 1 */
   wv1 = _inset[voice1][0];
   _spoke[54277] = _inset[voice1][1]; /* Set Attack/Decay for voice 1 */
   _spoke[54278] = _inset[voice1][2]; /* Set Sustain/Release for voice 1 */
   /* if a pulse wave, set pulse width */
   if (wv1 == 65) {
	  _spoke[54274] = _inset[voice1][3];/* low bits - pulse width for voice 1 */
	  _spoke[54275] = _inset[voice1][4];/* high bits - pulse width for voice 1 */
   }

   /* set-up voice 2 */
   if (voice2 > -1 ) {
	    wv2 = _inset[voice2][0];
   		_spoke[54284] = _inset[voice2][1];
   		_spoke[54285] = _inset[voice2][2];
   		if (wv2 == 65) {
   			_spoke[54281] = _inset[voice2][3];
			_spoke[54282] = _inset[voice2][4];
		}
	}
   /* set-up voice 3 */
   if (voice3 > -1) {
	    wv3 = _inset[voice3][0];
   		_spoke[54291] = _inset[voice3][1];
   		_spoke[54292] = _inset[voice3][2];
    	if (wv3 == 65) {
   			_spoke[54288] = _inset[voice3][3];
			_spoke[54289] = _inset[voice3][4];
		}

   }

   /* Set volume of at maximum - no filters */
   _spoke[54296] = 15; /* Volume and filter modes */
   idx = 0;

   while ((freq = song[idx]) != 0) {
	   idx++;
	   dur = song[idx];  idx++;

	   if (_spress[0] != 0x40)break; /* check for a key down */
       if (freq > 94) freq = 255;
       if (dur == 0)continue;

	   if (freq != 255) {
  		   _freq1[0] = _cfreq[freq];
		   _sid[4] = wv1; /* control register */

		   if (voice2 > -1 && (_hu == 0 || dur > 2)) {
			   _freq2[0] = _cfreq[freq+_hu];
			   _sid[11] = wv2;
		   }
		   if (voice3 > -1 && (_hu == 0 || dur > 2)) {
			   _freq3[0] = _cfreq[freq-_hu];
			   _sid[18] = wv3;
		   }

	   }

	   len = (dur * _tempo);        /* timing loop */
	   while (len > 0)len--;

	   if (freq != 255) {
	      _sid[4] = wv1-1;
	      if (voice2 > -1 && (_hu == 0 || dur > 2)) _sid[11] = wv2-1;
	      if (voice3 > -1 && (_hu == 0 || dur > 2)) _sid[18] = wv3-1;
	   }

   }

}


/* midi note number, duration in milliseconds */
/* valid instruments for this function are defined in the _inset table.
   there are 18 of them (0-17).

   while this is nowhere close to GM (general midi)
   this is probably adequate to provide quasi-musical sound effects
   within the waveforms in my presets.

   I have left an open slot (18) for a custom instrument
   which you must define properly if used since I do limited range
   checking.

   */
playnote(note,duration,voice1,voice2,voice3)
int note,duration,voice1,voice2,voice3;
{
	char song[4];


	duration = duration / 55; /* convert to 18.2 ticks per second */

	song[0] = note & 0xff;
	song[1] = duration & 0xff;
	song[2] = 0;
	song[3] = 0;

    play(song,voice1,voice2,voice3);

}


/* sets an instrument with new values */
/* set idx to -1 for custom */
/* use the tables above for reference */
setinst(idx, wv, a, d, s, r, lp, hp)
int idx, wv, a, d, s, r, lp, hp;
{

	int ad, sr;

	a = a & 0xf;
	d = d & 0xf;
	s = s & 0xf;
	r = r & 0xf;

	ad = (a * 16) + d; /* attack velocity */
	sr = (s * 16) + r; /* sustain velocity */


    if (idx < 0 || idx > 17)idx = 18; /* default custom instrument */

    switch(wv) {
		case 0:
		case 17: wv = 17; break; /* triangle */

		case 1:
		case 33: wv = 33; break; /* sawtooth */

		case 2:
		case 65: wv = 65; break; /* pulse */

		case 3:
		case 129:
		default:
		         wv = 129; break; /* noise */
	}

    _inset[idx][0] = wv;
    _inset[idx][1] = ad & 0xff; /* Set Attack/Decay */
    _inset[idx][2] = ad & 0xff; /* Set Sustain/Release */

    if (wv != 65) {
		lp = 0;
		hp = 0;
	}
	_inset[idx][3] = lp & 0xff; /* low bits - pulse width */
    _inset[idx][4] = hp & 0xf;  /* high bits - pulse width */

}

