/* KEYBRD.C (C) Copyright 2007 Bill Buckels */
/* retrieve C64 keypress values without echoing to the console */

/* This program demonstrates a strategy for capturing the C64 keyboard
including capturing the values of navigational keys in Aztec C.

While it is interesting, this is not the best way
that I have found to read the keyboard buffer directly on the C64.

The best way is to let the keyboard routines do their job
while the program runs. The buffer cannot hold more than
10 ascii values. The buffer is a FIFO queue. If the user
bangs on the keyboard more than 10 times before you clear it,
the buffer is full and additional keys are lost.

When Reading the buffer the best way is to:

1. Before Clearing the buffer wait until no key is being pressed.
2. Clear the buffer to 0's and set the buffer length to 0.
3. Wait until a key is being pressed.
4. Wait until no key is being pressed.
5. Get the ascii value in buffer[0]
6. Clear the buffer to 0's and set the buffer length to 0.

This is equivalent to a getch() function. Aztec C doesn't have
one and getchar() really doesn't do-it for me, especially
since a stupid "ball" prompt is displayed. getchar() is also
really annoying in graphics modes because it screws-up the
screen.

Variations of the above strategy are similar to the
kbhit() strategy in an IBM C program... which is
also not the same but similar to what I am calling kbcode() below.

The best way of all to check for a keypress in a loop
and such is to:

Clear the buffer before starting the loop:

1. Before Clearing the buffer wait until no key is being pressed.
2. Clear the buffer to 0's and set the buffer length to 0.

Start the loop and check for a keypress each time that you consider
appropriate:

3. Wait until a key is being pressed.
4. Break the loop and do any necessary clean-up.

Outside the loop your user may be impatiently pressing keys so
you need to wait for a release before clearing the buffer and
moving-on.

5. Wait until no key is being pressed.

After the loop you may wish to throw away the key value that is waiting.
If you do not then

5.1 Get the ascii value in buffer[0].

Whether you need the ascii value or not, the last step is still the
same:

6. Clear the buffer to 0's and set the buffer length to 0.

Using the ascii value in the buffer has the advantage over the
scan codes of having translated values for the whole C64 character
set which include ascii values for function keys F1-F8. Matrix code
translation is not required since the C64 did that for us.

Now as a final point, it may seem to you that it is a great deal
of effort to get a key press using the strategy that I have just
described. However, it is exactly the same strategy that I would
use in either a DOS or a WINDOWS program even today under some
circumstances.

All this code prevents "run-on" while you set-up parts of your program.

Back in the late '80's there was an engineer at Kraft, a manufacturer
of Mice and Joysticks, called Dave who helped me when I was having trouble
with mouse-clicks causing run-on, skipping over several screens and things
of that nature. The Microsoft examples that I had referred to until then
showed us acting on a button-down event which as computers got faster
and faster would remain down long after we loaded the next screen
and would therefore be captured ad-nauseum causing the next branch in
the code to be triggered in an unsolicited and often undesirable manner.

Dave's suggestion was so simple that I couldn't believe that it wasn't
common knowledge at the time... but it wasn't from what Microsoft was
putting-out in their devkits as samples...

He said to always wait for a button release after a button press before
trying to capture the mouse again. After receiving this valuable tip
my programs got really solid, and would stop where they were supposed
to stop, because I would act on a click, then wait for a release before
entering a control loop.

For a dragging event I would get my first datum from the press and my
last from the release. I also applied this same strategy to my key
events since keyboard and mouse needed to be synched. And then all
was clear.

Now I suppose that all this talk hasn't much to do with this program
because lots of what I said isn't demonstrated clearly here.

But my purpose in writing this program was to explore my options
before deciding my best strategy for aquiring keypress values
on a C64. At the end of the day, it turned-out that matrix codes
are pointless and just add extra overhead since they must be translated.

Ascii is what we want.

Now one final word of caution: In an Aztec C program, when we fool
around with the keyboard we really want to use getchar() at the
end of the program. Otherwise there are situations where Aztec C will
hang and not return cleanly to BASIC.

The nuances of this are application dependent and fairly involved
so I am not taking time to discuss all this right now but I have
provided working examples that exit in many ways for your review.

I also should mention that with a little creative effort you can
return an 9 character command to BASIC appended with a return
at the end of your program by stuffing the keyboard buffer.

C64 BASIC has mneumonics for its commands and can do something
meaningful with a cleverly organized buffer. But that's a
discussion for another day.


/*

203 Matrix code of key currently being pressed. Values:

$00-$3F: Keyboard matrix code.

$40: No key is currently pressed

*/


/*

198 Length of keyboard buffer. Values:

$00, 0: Buffer is empty.

$01-$0A, 1-10: Buffer length.

 */


 /*

 631-640 Keyboard buffer (10 bytes, 10 entries).

 */

/* the first 8 keyboard matrix codes are navigational keys */
#define DELKEY         0
#define RETKEY         1
#define LTRT           2
#define F7F8           3
#define F1F2           4
#define F3F4           5
#define F5F6           6
#define UPDN           7

/* the british pound sign is not ascii (of course) */
/* but it is petscii */
#define POUNDKEY       48

/* the last 16 bytes in the matrix are navigational keys
   mixed with character keys */
#define CLRHOME        51
#define UPARROW	       54
#define LTARROW        57
#define SPACEBAR       60
#define RUNKEY         63

/* I have encoded the matrix below with zero's
   where the key is not an ascii value.
   See the code below for how this is implemented. */
char kbchar[64] = {
  0x00,0x00,0x00,0x00,0x00,0x00,0x00,000,
  '3', 'W', 'A', '4', 'Z', 'S', 'E', '0',
  '5', 'R', 'D', '6', 'C', 'F', 'T', 'X',
  '7', 'Y', 'G', '8', 'B', 'H', 'U', 'V',
  '9', 'I', 'J', '0', 'M', 'K', 'O', 'N',
  '+', 'P', 'L', '-', '.', ':', '@', ',',
  0x00,'*', ';', 0x00,0x00,'=', 0x00,'/',
  '1', 0x00,0x00,'2', 0x00,0x00,'Q', 0x00};


char *keypress = (char *)203;
char *keybuf =  (char *)631;
char *keybuflen = (char *)198;

/* low level keyboard routine */
int kbcode()
{
	int ch, c;

	ch = 0x40;
	while (ch == 0x40)ch = keypress[0]; /* get keypress */
	c = ch;
	while (ch != 0x40)ch = keypress[0]; /* get key release */

	ch = 0;
	if (c < 0x40) {
	  ch = kbchar[c];
	}

    /* if character is navigational return scancode only */
    /* ch will be zero */
	return ((c << 8) | ch);

}

/* the keyboard must be flushed after using the above routine */
int kbflush()
{
	int c;

    c = keypress[0];

    if (c != 0x40) return 1;  /* keyboard is busy, try again later */

	for (c = 9; c > -1; c--) keybuf[c]=0;
	keybuflen[0] = 0;

	return 0;

}



main()
{

    int kb, c, ch;

    printf("%cC64 KEYBOARD MATRIX CODE DEMO\nPRESS X TO END\n",147);
    kb = keybuflen[0];
    printf("CURRENT BUFFER LENGTH IS %d\nVALUES: ",keybuflen[0]);
    for (c = 0; c < 10; c++) {
	   kb = keybuf[c];
	   printf("%d ",kb);
	}
	printf("\nPRESS ANY KEY...\n");
    ch = 0;
    while (ch != 'X') {

	    kb = kbcode();
	    ch = (kb & 0xff);
	    c  = (kb >> 8) & 0xff;

	    if (ch != 0) printf("KEYCODE IS %d - CHARACTER IS %c\n",c,ch);
        else printf("KEYCODE IS %d - NAVIGATION KEY\n",c);
        kb = keybuflen[0];
		printf("CURRENT BUFFER LENGTH IS %d\nVALUES: ",keybuflen[0]);
		for (c = 0; c < 10; c++) {
			   kb = keybuf[c];
			   printf("%d ",kb);
	    }
	    printf("\n");

    }


    printf("NOW CLEARING KEYBOARD BUFFER.\n");
    while (kbflush());
	kb = keybuflen[0];
	printf("CURRENT BUFFER LENGTH IS %d\nVALUES: ",keybuflen[0]);
	for (c = 0; c < 10; c++) {
	   kb = keybuf[c];
	   printf("%d ",kb);
	}
    printf("\nPRESS ANY KEY TO RETURN TO BASIC...");
    c= getchar();
    printf("%c",147);
    exit(0);

}
