/* piano.c (C) Copyright Bill Buckels 2007 */

/* This program is an Aztec C65 implementation of the pianokey SID
chip demonstration program from the Commodore 64 User's Guide 2nd
Edition available for download at

http://project64.c64.org/hw/c64.html

It has been built under Windows XP in the Aztec C65 cross-development
environment for the C64.

This program links with the standard native mode C64NAT.LIB provided
by Manx.

Like any C64 native mode program, this program is loaded from BASIC.
I have added startup and shutdown code so this program will exit
cleanly to BASIC when done. The original program just ran in a loop
with no way-out.

Consequently this is a somewhat less than faithful port of the
original example from the Commodore 64 User's Guide. Since the
original program had few comments I have conjured-up some of my own.

The waveforms are backwards in this program as compared to the C64
documentation and this program leaves a note playing until the next
one is selected which is the reason (I think) for this program doing
things backwards. I don't recommend doing "background" sound playback
this way but it certainly seems to work well for interactive
playback.

This program uses the 3 voice SID chip that the C64 sported. This
was a fairly complex bit of hardware and you should probably study
the Commodore 64 Programmer's Reference Guide if you require more
clarification.

However DO NOT use any frequency values in the C64 documentation,
nor even in the original piano key program for the chromatic scale
equivalents for notes. None of them with the exception I have noted
below are on-pitch when used with the WinVice emulator.

Every value including note-numbers in Commodore's C64 documentation
is wrong as far as I can see.

I generated my own frequency tables which are reasonably close to
being on pitch by using the frequency values for the 2 C notes from
the 1 octave scale in the C64 user's guide as a basis for
establishing the proper gap between C3 and C4.

+--------------+---------+---------------------------------------------------+
|TO PLAY A NOTE|      C  | C#| D | D#| E | F | F#| G | G#| A | A#| B | C | C#|
|HIGH FREQUENCY|54280 34 | 36| 38| 40| 43| 45| 48| 51| 54| 57| 61| 64| 68| 72|
|LOW FREQUENCY |54279 75 | 85|126|200| 52|198|127| 97|111|172|126|188|149|169|
+--------------+---------+------------+------------+------------+------------+

The values are still not right because they are off pitch but on key
(almost, read-on).

The pianokey program uses what seems to be an arbitrary value (but
probably isn't) to adjust the pitch between the higher frequency
octaves by doubling the frequency from the previous octave then
subtracting 30.

I think what it may be doing is the inverse of adding 15 to allow
for some bitwise underflow in a nibble that occurs maybe even in the
SID itself, which is more noticeable or only occurs in the higher
frequencies. I do know that the SID is not an FM chip like the
Yamaha OPL3. The SID actually creates a wide waveform and then
scales down the result somehow so I know that some info might be
lost, and might need to be corrected and I also know that Bob
Yannes, the SID's designer, was never given the time to document the
SID properly so the manuals probably suffered ever since and sre
still wrong in the actual values given. Some guys probably even know
the real match behind all this but I don't.

Rather than wear-myself-out wondering why pianokey's math is better
than the manual I used a simple scaling technique base on the value
given for C5 above which is on pitch at (68 * 256) + 149 = 17557.
You will see this value in the table below.

However when used with the other C value to create an octave, the key
was skewed. Since the key was accurate in pianokey I wondered what
was different. As it turned-out, the pianokey program used a value of
30 which was subtracted from the frequency value obtained to get the
next higher octave frequency after doubling the frequency which is
the method that I am familiar with and which is also the method
documented by commodore.

I did the inverse. I added 15 to the value obtained by taking the
lower frequency C4 in the table above, allowing that this value is
half the higher value rounded upwards.

(34 * 256) + 75 = 8779 + 15 = 8794

I then took pianokey's frequency table calculation method which uses
a seed value and changed it as follows in a GWBASIC program to
produce a frequency table for the piano program that is only
one-digit out from 8794. I ended-up with 8793 and this worked nicely
in this program. Here's the BASIC program for creating the corrected
table:

1  REM FR.BAS
5  REM GWBASIC program for generating on-pitch frequency table
6  REM for use in C64 piano program.
7  REM Copyright Bill Buckels 2008
10 DIM F(27)
20 F1=6403.5: FOR I=1 TO 26: F(27-I)=F1*5.8+30: F1=F1/2^(1/12): NEXT
25 OPEN "FR.TXT" for output as #1
26 PRINT #1, "int freq[27] = {"
27 PRINT #1, INT(0);","
30 FOR I=1 TO 25: PRINT #1, INT(F(I));",": NEXT I
31 PRINT #1, INT(F(26));"};"
35 CLOSE #1
40 SYSTEM


To recap, what this amounts to is a scaled algorithm from the
orginal pianokey program which was not on pitch but was on key. The
fact that the small table from the user's guide was close to
being-on pitch combined with the fact that the rest of the stuff
wasn't says to me that none of this stuff was ever right or may
have been partially right for 50hz machines or some earlier version
of the SID.

My frequencies are right (or at least are more correct).

Or perhaps I am wrong and pianokey was written for use in the WinVICE
emulator and maybe WinVICE has a bug in their SID emulation routines
since I have written this without even testing on a real C64.

You should also probably understand some of the physics of sound and
perhaps some musical theory, and maybe even some electrical
engineering to really grasp what is going on here. I have knowledge
gaps in this area and suffer from thinking I know-it-all but if you
aren't like me...

A good link to help you along is

http://en.wikipedia.org/wiki/6581

None of that is a pre-requisite to having some fun with this of
course as I obviously have.

*/


/* $D400-$D7FF, 54272-55295 SID; audio */
#define SID 54272

/* The following character table is a piano keyboard primarily
   using commodore 64 screen control and graphics characters.
   Since it was already as clear as mud I just threw-it into a
   0 terminated character array for the C version. The only
   man-readable characters are in single quotes. I am using
   white text instead of the usual light blue.
*/
char layout[] = {
147,5,32,18,32,29,32,29,32,194,32,29,32,29,
32,29,32,194,32,29,32,29,32,194,32,29,32,29,32,13,
32,18,32,29,32,29,32,194,32,29,32,29,32,29,32,194,
32,29,32,29,32,194,32,29,32,29,32,13,
32,18,32,29,32,29,32,194,32,29,32,29,32,29,32,194,
32,29,32,29,32,194,32,29,32,29,32,13,
32,18,32,194,32,194,32,194,32,194,32,194,32,194,32,
194,32,194,32,194,32,194,32,194,32,194,32,13,
32,18,'Q',194,'W',194,'E',194,'R',194,'T',194,'Y',194,'U',
194,'I',194,'O',194,'P',194,'@',194,'*',194,'^',13,
0};


/* the following keyboard matrix is a little wasteful
   memory wise but speeds-up frequency lookup values.
   it is indexed by ASCII (or more properly PETSCII).
   If a key value corresponds to a frequency the
   index in the frequency array is used to get the
   base frequency value. However, if the value is
   0 then the corresponding ASCII value may or may not
   be a control key or some other key.

   K$="Q2W3ER5T6Y7UI9O0P@-*\^"

*/

char fidx[256] = {
 0,
 0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,
 0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,
 0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,
 0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,
 0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,
 0 ,  20 ,  0 ,  0 ,  19 ,  0 ,  0 ,  16 ,
 0 ,  2 ,  4 ,  0 ,  7 ,  9 ,  11 ,  0 ,
 14 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  18 ,
 0 ,  0 ,  0 ,  0 ,  5 ,  0 ,  0 ,  0 ,
 13 ,  0 ,  0 ,  0 ,  0 ,  0 ,  15 ,  17 ,
 1 ,  6 ,  0 ,  8 ,  12 ,  0 ,  3 ,  0 ,
 10 ,  0 ,  0 ,  21 ,  0 ,  22 ,  0 ,  0 ,
 0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,
 0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,
 0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,
 0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,
 0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,
 0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,
 0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,
 0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,
 0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,
 0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,
 0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,
 0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,
 0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,
 0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,
 0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,
 0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,
 0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,
 0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,
 0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 ,
 0 ,  0 ,  0 ,  0 ,  0 ,  0 ,  0 };


/* the following frequency table is based on the chromatic scale */
/* the original program calculated these as floats */
/* I store them as a precalculated integer table. */
/* only values in the range of 1-22 are used in this program. */
/* it is a base 1 array like its BASIC counterpart. */
/* this does not match the original program except for the */
/* calculation method because I have adjusted this for pitch in the */
/* VICE emulator. See my rant above. */


 int freq[27] = {
	 0,
	 8793,
	 9315,
	 9867,
	 10452,
	 11071,
	 11728,
	 12424,
	 13161,
	 13941,
	 14769,
	 15645,
	 16574,
	 17557,
	 18600,
	 19704,
	 20874,
	 22113,
	 23426,
	 24818,
	 26292,
	 27853,
	 29508,
	 31261,
	 33118,
	 35085,
	 37170};

/* keyboard pointers */
char *key = (char *)203;
char *kbuf = (char *)631;
char *klen = (char *)198;

/* if you are going to poke around
   this is the best way that I have found */
char *poke = (char *)0;  /* a zero pointer */
char *sid = (char *)SID;

main()
{
	int ch, idx, jdx;
	int m, oc, w, wv;
	int at, de, su, re, sv, av;
    int fr, t, v, p;

    /* clear all keypressess on entry */
    while (key[0] != 0x40);
    /* clear the buffer */
    for (idx = 9; idx > -1; idx--) kbuf[idx]=0;
    klen[0] = 0;

    /* draw the piano keyboard */
	for (idx = 0; layout[idx] != 0; idx++)putchar(layout[idx]);
	/* display the menu options */
    printf("%c'NUMBER KEYS' ARE 'BLACK KEYS'\n",17);
    printf("%c'SPACE' FOR SOLO OR POLYPHONIC\n",17); /* polyphonic toggle */
    printf("%c'F1,F3,F5,F7' OCTAVE SELECTION\n",17);
    printf("%c'F2,F4,F6,F8' WAVEFORM\n\n",17);
    printf("%c PRESS 'X' TO END%c\n",17,17);

RESET:;

    for (idx = 0; idx < 25; idx++)sid[idx] = 0; /* clear sound chip */
    sid[4]  = 8;
	sid[4]  = 0; /* fix off */
    sid[11] = 8;
	sid[11] = 0; /* fix off */
    sid[18] = 8;
	sid[18] = 0; /* fix off */


    at = 0;   /* attack */
    de = 0;   /* decay  */
    su = 15;  /* sustain */
    re = 9;   /* release */
    sv = (su * 16) + re; /* sustain velocity */
    av = (at * 16) + de; /* attack velocity */

    m  = 1;   /* divisor for frequency table */
    oc = 4;   /* octave - this value is not actively used */
    w  = 0;   /* wave form */
    wv = 16;  /* wave velocity */

    /* setup 3 voices of SID chip */
    /* these were calculated values in the original program */
    /* you really need to study the memory map for the real poop */
    /* and you may want to review the basic program if you */
    /* can hold your nose long enough */
    /* voice 1 */
	poke[54277] = 0;  /* Attack and Decay length */
	poke[54278] = 249;/* Sustain and Release length */
	poke[54274] = 160;/* low bits - pulse width of 4000. */
	poke[54275] = 15; /* high bits - pulse width of 4000 */
    /* voice 2 */
	poke[54284] = 0;
	poke[54285] = 249;
	poke[54281] = 160;
	poke[54282] = 15;
    /* voice 3 */
	poke[54291] = 0;
	poke[54292] = 249;
	poke[54288] = 160;
	poke[54289] = 15;

    /* setup volume to full - no filters */
	poke[54296] = 15; /* Volume and filter modes */

    p = 0;
    v = 0;

    while (key[0] != 0x40); /* wait for key release */
    for (idx = 9; idx > -1; idx--) kbuf[idx]=0;
	klen[0] = 0;

TOP:;
    /* not waiting for a release - might be a little risky */
    while (klen[0] == 0);
    idx = klen[0];
    ch = kbuf[0];

    switch(idx) {
		case 0: break;
		case 1: klen[0] = 0; kbuf[0] = 0; break;
		case 2: klen[0] = 1; kbuf[0] = kbuf[1]; kbuf[1] = 0; break;
		case 3: klen[0] = 2; kbuf[0] = kbuf[1]; kbuf[1] = kbuf[2]; kbuf[2] = 0; break;
		default: klen[0] = idx-1;
                 for (idx =0, jdx = 1; idx < 9; idx++, jdx++) kbuf[idx]=kbuf[jdx];
                 kbuf[9] = 0;
                 break;
    }

    if (ch==0) goto TOP; /* safety */

	if (ch == 'X') goto BOTTOM;
	if (ch == 147) goto RESET;

   /* bit of a safety here */
   idx = fidx[ch];
   if (idx!=0) {
	 fr = freq[idx] / m;
	 t = v * 7;   /* 3 sets of 7 registers */

	 /* finish dec/sus */
	 sid[t+6] = 0;
	 /* finish att/rel */
	 sid[t+5] = 0;

	 sid[t+4] = 8;
	 sid[t+4] = 0; /* fix off */

	 sid[t] = (char)(fr & 0xff); /* set lo */;
	 sid[t+1] = (char)(fr >> 8); /* set hi */

	 /* set dec/sus */
	 sid[t+6] = sv;
	 /* set att/rel */
	 sid[t+5] = av;

	 sid[t+4] = wv+1; /* control register */
	 sid[t+4] = wv;   /* control register */

	 /* polyphonic so alternate between voices 0-3 */
	 if (p == 1) {
	   v++;
	   if (v > 2)v = 0; /* three voices in a sid chip */
	 }

   }
   else {
	      /* I am handling my control keys here */
		  /* F1,F3,F5,F7,F2,F4,F6,F8 - dumb old double keys */
		  switch(ch) {
			 case 133: m = 1; oc = 4; break;
			 case 134: m = 2; oc = 3; break;
			 case 135: m = 4; oc = 2; break;
			 case 136: m = 8; oc = 1; break;
			 case 137: w = 0; wv = 16;  break;
			 case 138: w = 1; wv = 32;  break;
			 case 139: w = 2; wv = 64;  break;
			 case 140: w = 3; wv = 128; break;
			 /* spacebar */
			 case 32:  if (p == 1) p = 0; /* poly toggle */
					   else (p = 1);
		  }

	}




goto TOP;


BOTTOM:;


    for (idx = 0; idx < 25; idx++)sid[idx] = 0; /* clear sound chip */
    sid[4]  = 8;
	sid[4]  = 0; /* fix off */
    sid[11] = 8;
	sid[11] = 0; /* fix off */
    sid[18] = 8;
	sid[18] = 0; /* fix off */

    putchar(154);
    /* clear all keypresses on exit */
    while (key[0] != 0x40); /* wait for release */
    /* clear the buffer */
    for (idx = 9; idx > 0; idx--) kbuf[idx]=0;
    kbuf[0] = 13; /* stuff the buffer - enter key */
    klen[0] = 1;
    getchar();

exit(0);
}